/***************************************************************************
                          texdat.c  -  description
                             -------------------
    begin                : Wed Jul 26 2000
    copyright            : (C) 2000 by Simon E. Reichelt
    email                : s.reichelt@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <gtk/gtk.h>

#include "wadfile.h"
#include "texdat.h"
#include "wadview.h"


byte_t *GetWAD2PalettePos(byte_t *ram_file)
{
	wadinfo_t w_info;
	lumpinfo_t l_info;
	
	memcpy(&w_info, ram_file, sizeof (wadinfo_t));
	ram_file += w_info.infotableofs;
	
	memcpy(&l_info, ram_file, sizeof (lumpinfo_t));
	ram_file = ram_file - w_info.infotableofs + l_info.filepos;
	
	return ram_file;
}	
	

byte_t *GetWAD3PalettePos(lumpinfo_t *l_info, byte_t *ram_file)
{
	byte_t *run_p;
	miptex_t miptex;
	
	run_p = ram_file + l_info->filepos;
	memcpy(&miptex, run_p, sizeof (miptex_t));
	
	run_p += miptex.offsets[3];
	run_p += (miptex.width / 8) * (miptex.height / 8);
	
	run_p += sizeof (short);
	
	return run_p;
}


GdkRgbCmap *GetColormap(byte_t *pal_pos)
{
	//unsigned int *c_pal;
	unsigned int c_pal[256];
	int i;
	unsigned char c_tripple[3];
	GdkRgbCmap *col_map;
	
	//c_pal = malloc(sizeof (unsigned int) * 256);
	
	for (i = 0; i < 256; i++)
	{
		memcpy(c_tripple, pal_pos, 3);
		
		c_pal[i] = (c_tripple[0] << 16) + (c_tripple[1] << 8) +
		            c_tripple[2];
		
		pal_pos += 3;
	}
	
	col_map = gdk_rgb_cmap_new(c_pal, 256);
	
	return col_map;
}


int GetMemOffset(int width, int height)
{
	int i, scale = 1;
	
	//Retrieve optimal mipmap level
		
	for (i = 0; i < 4; scale *= 2, i++)
		if ((width / scale) <= 64 && (height / scale) <= 64)
			return i;
			
	return 0;
}


void GetRealSize(texrep_t *tex, int *width, int *height)
{
	int off_info[4] = {1, 2, 4, 8};
	
	//Calculate the original tex size from
	//the mipmap level
	
	*width = tex->width * off_info[tex->off_s];
	*height = tex->height * off_info[tex->off_s];
}


int GetMiptexInfo(texrep_t *texture, miptex_t *miptex, int offset)
{
	int count;
	int tmp_width;
	int tmp_height;
	int off_info[4] = {1, 2, 4, 8};
	
	tmp_width = miptex->width / off_info[offset];
	tmp_height = miptex->height / off_info[offset];
	
	if (texture != NULL)
	{
		texture->width = tmp_width;
		texture->height = tmp_height;
	}	
		
	count = tmp_width * tmp_height;
	
	return count;
}


void FlushBackground(GdkWindow *xpm, GdkGC *gc, int width, int height)
{
	GdkColormap *clr_map;
	GdkColor rect_clr;
	
	//Set the texture image background
	//to a white filled rectangle
	
	clr_map = gdk_colormap_get_system();
	
	rect_clr.red = 35000;
	rect_clr.green = 35000;
	rect_clr.blue = 35000;
	
	gdk_colormap_alloc_color(clr_map, &rect_clr, FALSE, FALSE);
	
	gdk_gc_set_foreground(gc, &rect_clr);
	
	gdk_draw_rectangle(xpm, gc, TRUE, 0, 0, width, height);
}


void CorrectIndex(texrep_t *tex1, texrep_t *tex2)
{
	int tmp_index;
	
	tmp_index = tex1->index;
	tex1->index = tex2->index;
	tex2->index = tmp_index;
}


void SortCacheTextures(texrep_t *tex_set, int tex_cnt)
{
	int i, j, max_tex;	
	texrep_t tmp_tex;
	
	//Alphabetize texture list
	max_tex = tex_cnt;
	
	if (wad_type == WAD2)
		max_tex--;
	
	for (i = 0; i < max_tex; i++)
	{
		for (j = tex_cnt - 2; j > 0; j--)
		{
			if (strcasecmp(tex_set[j - 1].name,
			               tex_set[j].name) > 0)
			{
				memcpy(&tmp_tex, &tex_set[j - 1],
				       sizeof (texrep_t));
				memcpy(&tex_set[j - 1], &tex_set[j],
				       sizeof (texrep_t));
				memcpy(&tex_set[j], &tmp_tex,
				       sizeof (texrep_t));
				
				CorrectIndex(&tex_set[j - 1],
				             &tex_set[j]);
			}
		}
	}
}


byte_t *DrawCmapPixmap(void)
{
	int i, j, k, l;
	byte_t c_pos = 0;
	byte_t *m_pos;
	
	//Create the colormap visualisation
	//pixmap at the end of the texture cache
	
	m_pos = malloc(49152 * sizeof (char));
	
	for (i = 0; i < 16; i++)
	{
		for (j = 0; j < 12; j++)
		{
			for (k = 0; k < 16; k++)
			{
				for (l = 0; l < 16; l++)
					*m_pos++ = c_pos + k;
			}
		}
		
		c_pos += 16;
	}
	
	return (m_pos - 49152);
}


void SetCmapValues(texrep_t *tex_entry)
{
	strcpy(tex_entry->name, "colormap");
	
	tex_entry->zero_offs = NULL;
	tex_entry->d_area = NULL;
}


void RenderMipmap(byte_t **d_pxm, byte_t *tex_read, int offset)
{
	miptex_t miptex;
	texrep_t tex_entry;
	int count;
	int width, height;
	
	//Rasterize a mipmap pixmap
	
	memcpy(&miptex, tex_read, sizeof (miptex_t));
	
	tex_read += miptex.offsets[offset];
	
	count = GetMiptexInfo(&tex_entry, &miptex, offset);
	
	if (tex_entry.width <= 64 &&
	    tex_entry.height <= 64)
		width = height = 64;
	else
	{
		width = miptex.width;
		height = miptex.height;
	}
	
	*d_pxm = malloc(count * sizeof (char));
	if (!d_pxm)
	{
		WVError("Ran out of memory!");
		return;
	}
	
	memcpy(*d_pxm, tex_read, count * sizeof (char));
}


texrep_t *CacheWadTextures(FILE *w_file, GdkGC *gc)
{
	texrep_t *textures;
	byte_t *ram_file, *lmp_pos;
	byte_t *tex_read;
	wadinfo_t w_info;
	miptex_t miptex;
	lumpinfo_t lmp_info;
	int i, m_offset, max_tex;	
	GdkRgbCmap *g_cmap = NULL;
	
	//Create the texture pixmap array
	
	ram_file = LoadFileFromDisk(w_file);
	
	if (wad_type == WAD2)
		g_cmap = GetColormap(GetWAD2PalettePos(ram_file));
		
	memcpy(&w_info, ram_file, sizeof (wadinfo_t));
	
	textures = malloc((w_info.numlumps + 1) * sizeof (texrep_t));
	if (CheckMemory((void*) textures) == BAD_MEM ||
	    CheckMemory((void*) ram_file) == BAD_MEM)
	{
		free(ram_file);
		return NULL;
	}
	
	lmp_pos = ram_file + w_info.infotableofs;
	if (wad_type == WAD2)
		lmp_pos += sizeof (lumpinfo_t);
	
	max_tex = w_info.numlumps;
	if (wad_type == WAD2)
		max_tex--;	
	
	for (i = 0; i < max_tex; i++)
	{
		memcpy(&lmp_info, lmp_pos, sizeof (lumpinfo_t));
		//strcpy(textures[i].name, lmp_info.name);
		
		if (lmp_info.type != MIPTEX(wad_type))
		{
			ConPrint("CacheWadTextures(): WAD file isn't a texture wad.\n");
			free(ram_file);					
			return 0;
		}
		
		tex_read = ram_file + lmp_info.filepos;
		memcpy(&miptex, tex_read, sizeof (miptex_t));
		
		m_offset = GetMemOffset(miptex.width, miptex.height);
		
		textures[i].index = i;
		textures[i].off_s = m_offset;
		strcpy(textures[i].name, miptex.name);
		
		if (wad_type == WAD2)
			textures[i].clr_map = g_cmap;
		else if (wad_type == WAD3)
			textures[i].clr_map = GetColormap(
			                      GetWAD3PalettePos(&lmp_info, ram_file));
		
		GetMiptexInfo(&textures[i], &miptex, m_offset);
		
		RenderMipmap(&textures[i].pxm_data, tex_read, m_offset);
		
		if (m_offset != 0)
			RenderMipmap(&textures[i].zero_offs, tex_read, 0);
		else
			textures[i].zero_offs = NULL;
		
		lmp_pos += sizeof (lumpinfo_t);
	}
	
	if (wad_type == WAD2)
	{
		textures[w_info.numlumps - 1].pxm_data = DrawCmapPixmap();
	 	SetCmapValues(&textures[w_info.numlumps - 1]);
	
		textures[w_info.numlumps - 1].clr_map = g_cmap;
	}
	
	free(ram_file);
	
	return textures;
}


int RedrawTex(GtkWidget *emitter, GdkEvent *event, gpointer data)
{
	GdkGC *gc;
	texrep_t *tex_entry;
	
	//Expose Event handler for texture entries
	
	gc = gdk_gc_new(emitter->window);
	
	tex_entry = (texrep_t*) data;
	
	FlushBackground(tex_entry->d_area->window, gc, tex_entry->width, 64);	
	
	gdk_draw_indexed_image(tex_entry->d_area->window, gc, 0, 0, tex_entry->width,
	                       tex_entry->height, GDK_RGB_DITHER_NORMAL,
	                       tex_entry->pxm_data, tex_entry->width,
	                       tex_entry->clr_map);
	
	if (tex_entry->index == sel_index)
	{
		PrepareRectGC(gc);
		
		gdk_draw_rectangle(tex_entry->d_area->window, gc, FALSE, 2, 2,
	                       tex_entry->width - 4, tex_entry->height - 4);
	}		
	
	return TRUE;
}


void FreeTextures(texrep_t *tex_pointer)
{
	int i;
	
	//Free the memory allocated by the
	//texture representations
	
	for (i = 0; i < tex_cnt; i++)
	{
		free(tex_pointer[i].pxm_data);
		
		if (tex_pointer[i].zero_offs != NULL)
			free(tex_pointer[i].zero_offs);
			
		if (wad_type == WAD3)
			gdk_rgb_cmap_free(tex_pointer[i].clr_map);
	}
	
	if (wad_type == WAD2)
		gdk_rgb_cmap_free(tex_pointer[0].clr_map);
	
	free(tex_pointer);	
}


short CorrectShort(short short_val)
{
	char short_t[2];
	
	memcpy(short_t, &short_val, 2);
	
	return ((short_t[0] << 8) + short_t[1]);
}


long CorrectLong(long long_val)
{
	char long_t[4];
	
	memcpy(long_t, &long_val, 4);
	
	return ((long_t[0] << 24) + (long_t[1] << 16) +
	        (long_t[2] << 8) + long_t[3]);
}


void WriteSGIHeader(FILE *sgi_file, texrep_t *tex_entry)
{
	sgi_header_t f_header;
	int width, height;
	
	GetRealSize(tex_entry, &width, &height);
	
	f_header.magic_n = CorrectShort(474);
	f_header.storage = FALSE;
	f_header.bpc = 1;
	f_header.dimension = CorrectShort(3);
	f_header.x_size = CorrectShort(width);
	f_header.y_size = CorrectShort(height);
	f_header.z_size = CorrectShort(3);
	f_header.pix_min = CorrectLong(0);
	f_header.pix_max = CorrectLong(255);
	strcpy(f_header.img_name, tex_entry->name);
	f_header.colormap = CorrectLong(0);
	
	fwrite(&f_header, sizeof (sgi_header_t), 1, sgi_file);
}


void WriteRGBData(FILE *sgi_file, texrep_t *tex_entry)
{
	int width, height;
	byte_t *rgb_buff, *tex_dat;
	int i, j = 0, b_pos = 0, k;
	int c_pos = 0;
	byte_t colors[768];
	unsigned int color;
	
	GetRealSize(tex_entry, &width, &height);
	
	rgb_buff = malloc((width * height) * 3);
	
	if (!tex_entry->zero_offs)
		tex_dat = tex_entry->pxm_data;
	else
		tex_dat = tex_entry->zero_offs;
	
	for (i = 0; i < 256; i++)
	{
		color = tex_entry->clr_map->colors[i];
		
		colors[j++] = color >> 16;		
		colors[j++] = (byte_t) ((color & 0xFFFF) >> 8);		
		colors[j++] = (byte_t) color & 0xFF;
	}
	
	b_pos = (width * height) * 3;
	b_pos += width - 2;
				
	for (j = 2; j >= 0; j--)
	{
		c_pos = 0;
		
		for (i = 0; i < height; i++)
		{
			b_pos -= (2 * width);
			
			for (k = 0; k < width; k++, c_pos++)
				rgb_buff[b_pos++ + j] = colors[(tex_dat[c_pos] * 3) + j];
		}
		
		b_pos++;
	}
			
	fwrite(rgb_buff, (width * height) * 3, 1, sgi_file);
	
	free(rgb_buff);	
}


int WriteSGIFile(char f_name[], texrep_t *tex_entry)
{
	FILE *sgi_file;
	
	sgi_file = fopen(f_name, "w");
	if (!sgi_file)
	{
		WVError("Couldn't Create File!");
		return -1;
	}
	
	WriteSGIHeader(sgi_file, tex_entry);
	
	WriteRGBData(sgi_file, tex_entry);
	
	fclose(sgi_file);
	
	ConPrint("Texture \"%s\" has been exported.\n",
	         tex_entry->name);
	
	return 0;
}


void TruncateTextureName(char target_pos[], char tex_name[])
{
	strcpy(target_pos, tex_name);
	
	if (strlen(tex_name) <= 9)
		return;
	else
	{		
		target_pos[6] = '\0';
		
		strcat(target_pos, "...");
	}
}